// Implements the data sets and assignment properties for item set bonuses (additional bonuses
// when multiple pieces of an equipment set are worn by a player).

#ifndef ITEMSET_H
#define ITEMSET_H

#include <string>
#include <vector>
#include <map>
#include <filesystem>

#include "Creature.h"
#include "ScriptCore.h"

class CreatureInstance;
class ReportBuffer;
class ItemSetManager;

using namespace std;
namespace fs = filesystem;

class ItemSetData
{
public:
	string mSetName;        //The set name.  Also used as the script name to call.  Interpreted as a script label to call.  Must not contain spaces.
	int mMinItem;                //The player must equip at least this many items in order for any set bonuses to be examined and applied. 
	string mItemList;       //A comma separated list of ItemDef IDs that belong to this set.
	string mFlavorText;     //Optional flavor text to append to an item that belongs to a set.

	ItemSetData();
	void Clear(void);
	void CopyFrom(const ItemSetData &source);
};


namespace ItemSetScript
{
	
enum ItemSetScriptExtOpCodes
{
	OP_ADD = OP_MAXCOREOPCODE,
	OP_ADD_F,
};


class ItemSetScriptDef : public ScriptCore::ScriptDef
{
	friend class ItemSetManager;  //Need to expose the script to allow advanced debugging like label lookups and instruction scans.
private:
	virtual void GetExtendedOpCodeTable(OpCodeInfo **arrayStart, size_t &arraySize);
};

class ItemSetScriptPlayer : public ScriptCore::ScriptPlayer
{
	virtual void RunImplementationCommands(int opcode);

public:
	CreatureInstance *attachedCreature;
	void DebugGenerateReport(ReportBuffer &report);
};

} //namespace ItemSet

struct ItemSetTally
{
	map<int, int> mItemIDCount;           //Tallies how many times an item has been counted.  This is to prevent dual rings from counting as two equipped set items.
	map<string, int> mSetNameCount;  //Counts how many items are equipped that belong to a particular set.  Does not count multiples (such as rings).
	void TallyItem(int itemID, const string &setName);
};



// Stores which items are used in which sets.
class ItemSetManager
{
public:
	void LoadData(void);                                //Load all item set data.
	void CheckItem(int itemID, ItemSetTally &output);  //Check if the item belongs to a set.  If it does, increment the output tally.
	void UpdateCreatureBonuses(ItemSetTally &tally, CreatureInstance *actor);  //Scan the item set tally, and apply any bonuses if applicable.

private:
	map<int, string> mRegisteredItemID;        //Maps item IDs (int) to a script name (string).
	map<string, ItemSetData> mRegisteredSets;  //Maps set names to the information of a specific set.
	ItemSetScript::ItemSetScriptDef mScriptDef;
	
	void LoadFile(const fs::path &filename);                 //Loads a table file defining sets to be registered.
	void UpdateFlavorText(void);
	void RegisterSet(const ItemSetData &data);
	void RegisterItem(int itemID, const string &setName);  //Register an item into a set.
	const ItemSetData* GetSetData(const string &setName);
};

extern ItemSetManager g_ItemSetManager;

#endif //#ifndef ITEMSET_H
